import logging

from databricks.labs.remorph.helpers.execution_time import timeit

logger = logging.getLogger(__name__)


@timeit
def recon(recon_conf, conn_profile, source, report):
    logger.info(recon_conf)
    logger.info(conn_profile)
    logger.info(source)
    logger.info(report)
    pass



import logging
from io import StringIO

from databricks.connect import DatabricksSession
from databricks.sdk.core import Config
from pyspark.sql.utils import AnalysisException, ParseException

from databricks.labs.remorph.config import MorphConfig

logger = logging.getLogger(__name__)


class Validate:
    """
    The Validate class is used to validate SQL queries using a Spark session.

    Attributes:
    - spark (SparkSession): The Spark session used to execute and validate SQL queries.
    """

    def __init__(self, sdk_config: Config):
        """
        Initialize the Validate class with Sparksession.
        """

        self.spark = DatabricksSession.builder.sdkConfig(sdk_config).getOrCreate()

    def validate_format_result(self, config: MorphConfig, input_sql: str):
        """
        Validates the SQL query and formats the result.

        This function validates the SQL query based on the provided configuration. If the query is valid,
        it appends a semicolon to the end of the query. If the query is not valid, it formats the error message.

        Parameters:
        - config (MorphConfig): The configuration for the validation.
        - input_sql (str): The SQL query to be validated.

        Returns:
        - tuple: A tuple containing the result of the validation and the exception message (if any).
        """
        catalog_name = config.catalog_name
        schema_name = config.schema_name
        logger.debug(f"Validation query with catalog {catalog_name} and schema {schema_name}")
        (flag, exception) = self.query(input_sql, catalog_name, schema_name)
        if flag:
            result = input_sql + "\n;\n"
            exception = None
        else:
            query = ""
            if "[UNRESOLVED_ROUTINE]" in exception:
                query = input_sql
            buffer = StringIO()
            buffer.write("-------------- Exception Start-------------------\n")
            buffer.write("/* \n")
            buffer.write(exception)
            buffer.write("\n */ \n")
            buffer.write(query)
            buffer.write("\n ---------------Exception End --------------------\n")

            result = buffer.getvalue()

        return result, exception

    # [TODO] Implement Debugger Logger
    def query(self, query: str, catalog_name=None, schema_name=None):
        """
        Validate a given SQL query using the Spark session.

        Parameters:
        - query (str): The SQL query to be validated.
        - catalog_nm (str): The catalog name for the Spark session.
        - schema_nm (str): The schema name for the Spark session.

        Returns:
        - tuple: A tuple containing a boolean indicating whether the query is valid or not,
        and a string containing a success message or an exception message.
        """
        spark = self.spark

        try:
            # [TODO]: Explain needs to redirected to different console
            # [TODO]: Hacky way to replace variables representation
            if catalog_name in (None, "transpiler_test") and schema_name in (None, "convertor_test"):
                spark.sql("create catalog if not exists transpiler_test")
                spark.sql("use catalog transpiler_test")
                spark.sql("create schema if not exists convertor_test")
                spark.sql("use convertor_test")
            else:
                spark.sql(f"use catalog {catalog_name}")
                spark.sql(f"use {schema_name}")

            # When variables is mentioned Explain fails we need way to replace them before explain is executed.
            spark.sql(query.replace("${", "`{").replace("}", "}`").replace("``", "`")).explain(True)
            return True, None
        except ParseException as pe:
            logger.debug("Syntax Exception : NOT IGNORED. Flag as syntax error :" + str(pe))
            return False, str(pe)
        except AnalysisException as aex:
            if "[TABLE_OR_VIEW_NOT_FOUND]" in str(aex):
                logger.debug("Analysis Exception : IGNORED: " + str(aex))
                return True, str(aex)
            if "[TABLE_OR_VIEW_ALREADY_EXISTS]" in str(aex):
                logger.debug("Analysis Exception : IGNORED: " + str(aex))
                return True, str(aex)
            elif "[UNRESOLVED_ROUTINE]" in str(aex):
                logger.debug("Analysis Exception : NOT IGNORED: Flag as Function Missing error" + str(aex))
                return False, str(aex)
            elif "Hive support is required to CREATE Hive TABLE (AS SELECT).;" in str(aex):
                logger.debug("Analysis Exception : IGNORED: " + str(aex))
                return True, str(aex)
            else:
                logger.debug("Unknown Exception: " + str(aex))
                return False, str(aex)
        except Exception as e:
            logger.debug("Other Exception : NOT IGNORED. Flagged :" + str(e))
            return False, str(e)



import copy
import logging
import re
import typing as t
from typing import ClassVar

from sqlglot import exp, parser
from sqlglot.dialects.dialect import parse_date_delta
from sqlglot.dialects.snowflake import Snowflake
from sqlglot.dialects.snowflake import _parse_to_timestamp as parse_to_timestamp
from sqlglot.errors import ParseError
from sqlglot.helper import seq_get
from sqlglot.tokens import Token, TokenType
from sqlglot.trie import new_trie

from databricks.labs.remorph.snow import local_expression

logger = logging.getLogger(__name__)

""" SF Supported Date and Time Parts:
    https://docs.snowflake.com/en/sql-reference/functions-date-time#label-supported-date-time-parts
    Covers DATEADD, DATEDIFF, DATE_TRUNC, LAST_DAY
"""
DATE_DELTA_INTERVAL = {
    "years": "year",
    "year": "year",
    "yrs": "year",
    "yr": "year",
    "yyyy": "year",
    "yyy": "year",
    "yy": "year",
    "y": "year",
    "quarters": "quarter",
    "quarter": "quarter",
    "qtrs": "quarter",
    "qtr": "quarter",
    "q": "quarter",
    "months": "month",
    "month": "month",
    "mons": "month",
    "mon": "month",
    "mm": "month",
    "weekofyear": "week",
    "week": "week",
    "woy": "week",
    "wy": "week",
    "wk": "week",
    "w": "week",
    "dayofmonth": "day",
    "days": "day",
    "day": "day",
    "dd": "day",
    "d": "day",
}


def _parse_dateadd(args: list) -> exp.DateAdd:
    return exp.DateAdd(this=seq_get(args, 2), expression=seq_get(args, 1), unit=seq_get(args, 0))


def _parse_split_part(args: list) -> local_expression.SplitPart:
    if len(args) != 3:
        err_msg = f"Error Parsing args `{args}`. Number of args must be 3, given {len(args)}"
        raise ParseError(err_msg)
    part_num = seq_get(args, 2)
    if isinstance(part_num, exp.Literal):
        # In Snowflake if the partNumber is 0, it is treated as 1.
        # Please refer to https://docs.snowflake.com/en/sql-reference/functions/split_part
        if part_num.is_int and int(part_num.name) == 0:
            part_num = exp.Literal.number(1)
    else:
        cond = exp.EQ(this=part_num, expression=exp.Literal.number(0))
        part_num = exp.If(this=cond, true=exp.Literal.number(1), false=part_num)
    return local_expression.SplitPart(this=seq_get(args, 0), expression=seq_get(args, 1), partNum=part_num)


def _div0null_to_if(args: list) -> exp.If:
    cond = exp.Or(
        this=exp.EQ(this=seq_get(args, 1), expression=exp.Literal.number(0)),
        expression=exp.Is(this=seq_get(args, 1), expression=exp.Null()),
    )
    true = exp.Literal.number(0)
    false = exp.Div(this=seq_get(args, 0), expression=seq_get(args, 1))
    return exp.If(this=cond, true=true, false=false)


def _parse_json_extract_path_text(args: list) -> local_expression.JsonExtractPathText:
    if len(args) != 2:
        err_message = f"Error Parsing args `{args}`. Number of args must be 2, given {len(args)}"
        raise ParseError(err_message)
    return local_expression.JsonExtractPathText(this=seq_get(args, 0), path_name=seq_get(args, 1))


def _parse_array_contains(args: list) -> exp.ArrayContains:
    if len(args) != 2:
        err_message = f"Error Parsing args `{args}`. Number of args must be 2, given {len(args)}"
        raise ParseError(err_message)
    return exp.ArrayContains(this=seq_get(args, 1), expression=seq_get(args, 0))


def _parse_dayname(args: list) -> local_expression.DateFormat:
    """
        * E, EE, EEE, returns short day name (Mon)
        * EEEE, returns full day name (Monday)
    :param args: node expression
    :return: DateFormat with `E` format
    """
    if len(args) != 1:
        err_message = f"Error Parsing args `{args}`. Number of args must be 1, given {len(args)}"
        raise ParseError(err_message)
    return local_expression.DateFormat(this=seq_get(args, 0), expression=exp.Literal.string("E"))


def _parse_trytonumber(args: list) -> local_expression.TryToNumber:
    if len(args) == 1 or len(args) == 3:
        msg = f"""Error Parsing args `{args}`:
                             * `format` is required
                             * `precision` and `scale` both are required [if specified]
                          """
        raise ParseError(msg)

    if len(args) == 4:
        return local_expression.TryToNumber(
            this=seq_get(args, 0), expression=seq_get(args, 1), precision=seq_get(args, 2), scale=seq_get(args, 3)
        )

    return local_expression.TryToNumber(this=seq_get(args, 0), expression=seq_get(args, 1))


def _parse_monthname(args: list) -> local_expression.DateFormat:
    if len(args) != 1:
        err_message = f"Error Parsing args `{args}`. Number of args must be 1, given {len(args)}"
        raise ParseError(err_message)
    return local_expression.DateFormat(this=seq_get(args, 0), expression=exp.Literal.string("MMM"))


def _parse_object_construct(args: list) -> exp.StarMap | exp.Struct:
    expression = parser.parse_var_map(args)

    if isinstance(expression, exp.StarMap):
        return exp.Struct(expressions=[expression.this])

    return exp.Struct(
        expressions=[t.cast(exp.Condition, k).eq(v) for k, v in zip(expression.keys, expression.values, strict=False)]
    )


def _parse_to_boolean(args: list, *, error=False) -> local_expression.ToBoolean:
    this_arg = seq_get(args, 0)
    return local_expression.ToBoolean(this=this_arg, raise_error=exp.Literal.number(1 if error else 0))


def _parse_tonumber(args: list) -> local_expression.ToNumber:
    if len(args) > 4:
        error_msg = f"""Error Parsing args args:
                          * Number of args cannot be more than `4`, given `{len(args)}`
                          """
        raise ParseError(error_msg)

    if len(args) == 1:
        return local_expression.ToNumber(this=seq_get(args, 0))
    elif len(args) == 3:
        return local_expression.ToNumber(this=seq_get(args, 0), precision=seq_get(args, 1), scale=seq_get(args, 2))
    elif len(args) == 4:
        return local_expression.ToNumber(
            this=seq_get(args, 0), expression=seq_get(args, 1), precision=seq_get(args, 2), scale=seq_get(args, 3)
        )

    return local_expression.ToNumber(this=seq_get(args, 0), expression=seq_get(args, 1))


class Snow(Snowflake):
    # Instantiate Snowflake Dialect
    snowflake = Snowflake()

    class Tokenizer(snowflake.Tokenizer):
        IDENTIFIERS: ClassVar[list[str]] = ['"']

        COMMENTS: ClassVar[list[str]] = ["--", "//", ("/*", "*/")]
        STRING_ESCAPES: ClassVar[list[str]] = ["\\", "'"]

        CUSTOM_TOKEN_MAP: ClassVar[dict] = {
            r"(?i)CREATE\s+OR\s+REPLACE\s+PROCEDURE": TokenType.PROCEDURE,
            r"(?i)var\s+\w+\s+=\s+\w+?": TokenType.VAR,
        }

        KEYWORDS: ClassVar[dict] = {**Snowflake.Tokenizer.KEYWORDS}
        # DEC is not a reserved keyword in Snowflake it can be used as table alias
        KEYWORDS.pop("DEC")

        @classmethod
        def update_keywords(cls, new_key_word_dict):
            cls.KEYWORDS = new_key_word_dict | cls.KEYWORDS

        @classmethod
        def merge_trie(cls, parent_trie, new_trie):
            merged_trie = {}
            logger.debug(f"The Parent Trie is {parent_trie}")
            logger.debug(f"The Input Trie is {new_trie}")
            for key in set(parent_trie.keys()) | set(new_trie.keys()):  # Get all unique keys from both tries
                if key in parent_trie and key in new_trie:  # If the key is in both tries, merge the subtries
                    if isinstance(parent_trie[key], dict) and isinstance(new_trie[key], dict):
                        logger.debug(f"New trie inside the key is {new_trie}")
                        logger.debug(f"Parent trie inside the key is {parent_trie}")
                        merged_trie[key] = cls.merge_trie(parent_trie[key], new_trie[key])
                        logger.debug(f"Merged Trie is {merged_trie}")
                    elif isinstance(parent_trie[key], dict):
                        merged_trie[key] = parent_trie[key]
                    else:
                        merged_trie[key] = new_trie[key]
                elif key in parent_trie:  # If the key is only in trie1, add it to the merged trie
                    merged_trie[key] = parent_trie[key]
                else:  # If the key is only in trie2, add it to the merged trie
                    merged_trie[key] = new_trie[key]
            return merged_trie

        @classmethod
        def update_keyword_trie(
            cls,
            new_trie,
            parent_trie=None,
        ):
            if parent_trie is None:
                parent_trie = cls._KEYWORD_TRIE
            cls.KEYWORD_TRIE = cls.merge_trie(parent_trie, new_trie)

        def match_strings_token_dict(self, string, pattern_dict):
            result_dict = {}
            for pattern in pattern_dict:
                matches = re.finditer(pattern, string, re.MULTILINE | re.IGNORECASE | re.DOTALL)
                for _, match in enumerate(matches, start=1):
                    result_dict[match.group().upper()] = pattern_dict[pattern]
            return result_dict

        def match_strings_list(self, string, pattern_dict):
            result = []
            for pattern in pattern_dict:
                matches = re.finditer(pattern, string, re.MULTILINE | re.IGNORECASE | re.DOTALL)
                for _, match in enumerate(matches, start=1):
                    result.append(match.group().upper())
            return result

        def tokenize(self, sql: str) -> list[Token]:
            """Returns a list of tokens corresponding to the SQL string `sql`."""
            self.reset()
            self.sql = sql
            # Update Keywords
            ref_dict = self.match_strings_token_dict(sql, self.CUSTOM_TOKEN_MAP)
            self.update_keywords(ref_dict)
            # Update Keyword Trie
            custom_trie = new_trie(self.match_strings_list(sql, self.CUSTOM_TOKEN_MAP))
            logger.debug(
                f"The New Trie after adding the REF, VAR and IF ELSE blocks "
                f"based on {self.CUSTOM_TOKEN_MAP}, is \n\n {custom_trie}"
            )
            self.update_keyword_trie(custom_trie)
            logger.debug(f"Updated New Trie is {self.KEYWORD_TRIE}")
            # Parent Code
            self.size = len(sql)
            try:
                self._scan()
            except Exception as e:
                start = self._current - 50
                end = self._current + 50
                start = start if start > 0 else 0
                end = end if end < self.size else self.size - 1
                context = self.sql[start:end]
                msg = f"Error tokenizing '{context}'"
                raise ParseError(msg) from e
            return self.tokens

    class Parser(snowflake.Parser):
        FUNCTIONS: ClassVar[dict] = {
            **Snowflake.Parser.FUNCTIONS,
            "ARRAY_AGG": exp.ArrayAgg.from_arg_list,
            "STRTOK_TO_ARRAY": local_expression.Split.from_arg_list,
            "DATE_FROM_PARTS": local_expression.MakeDate.from_arg_list,
            "CONVERT_TIMEZONE": local_expression.ConvertTimeZone.from_arg_list,
            "TRY_TO_DATE": local_expression.TryToDate.from_arg_list,
            "STRTOK": local_expression.StrTok.from_arg_list,
            "SPLIT_PART": _parse_split_part,
            "TIMESTAMPADD": _parse_dateadd,
            "TRY_TO_DECIMAL": _parse_trytonumber,
            "TRY_TO_NUMBER": _parse_trytonumber,
            "TRY_TO_NUMERIC": _parse_trytonumber,
            "DATEADD": parse_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
            "DATEDIFF": parse_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
            "IS_INTEGER": local_expression.IsInteger.from_arg_list,
            "DIV0NULL": _div0null_to_if,
            "JSON_EXTRACT_PATH_TEXT": _parse_json_extract_path_text,
            "BITOR_AGG": local_expression.BitOr.from_arg_list,
            "ARRAY_CONTAINS": _parse_array_contains,
            "DAYNAME": _parse_dayname,
            "BASE64_ENCODE": exp.ToBase64.from_arg_list,
            "BASE64_DECODE_STRING": exp.FromBase64.from_arg_list,
            "TRY_BASE64_DECODE_STRING": exp.FromBase64.from_arg_list,
            "ARRAY_CONSTRUCT_COMPACT": local_expression.ArrayConstructCompact.from_arg_list,
            "ARRAY_INTERSECTION": local_expression.ArrayIntersection.from_arg_list,
            "ARRAY_SLICE": local_expression.ArraySlice.from_arg_list,
            "MONTHNAME": _parse_monthname,
            "MONTH_NAME": _parse_monthname,
            "OBJECT_CONSTRUCT": _parse_object_construct,
            "OBJECT_KEYS": local_expression.ObjectKeys.from_arg_list,
            "TRY_PARSE_JSON": exp.ParseJSON.from_arg_list,
            "TIMEDIFF": parse_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
            "TIMESTAMPDIFF": parse_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
            "TIMEADD": _parse_dateadd,
            "TO_BOOLEAN": lambda args: _parse_to_boolean(args, error=True),
            "TO_DECIMAL": _parse_tonumber,
            "TO_DOUBLE": local_expression.ToDouble.from_arg_list,
            "TO_NUMBER": _parse_tonumber,
            "TO_NUMERIC": _parse_tonumber,
            "TO_OBJECT": local_expression.ToObject.from_arg_list,
            "TO_TIME": parse_to_timestamp,
            "TIMESTAMP_FROM_PARTS": local_expression.TimestampFromParts.from_arg_list,
            "TO_VARIANT": local_expression.ToVariant.from_arg_list,
            "TRY_TO_BOOLEAN": lambda args: _parse_to_boolean(args, error=False),
            "UUID_STRING": local_expression.UUID.from_arg_list,
            "SYSDATE": exp.CurrentTimestamp.from_arg_list,
            "TRUNC": lambda args: local_expression.DateTrunc(unit=seq_get(args, 1), this=seq_get(args, 0)),
            "APPROX_PERCENTILE": exp.ApproxQuantile.from_arg_list,
            "NTH_VALUE": local_expression.NthValue.from_arg_list,
        }

        FUNCTION_PARSERS: ClassVar[dict] = {
            **Snowflake.Parser.FUNCTION_PARSERS,
            "LISTAGG": lambda self: self._parse_list_agg(),
        }

        PLACEHOLDER_PARSERS: ClassVar[dict] = {
            **Snowflake.Parser.PLACEHOLDER_PARSERS,
            TokenType.PARAMETER: lambda self: self._parse_parameter(),
        }

        FUNC_TOKENS: ClassVar[dict] = {*Snowflake.Parser.FUNC_TOKENS, TokenType.COLLATE}

        COLUMN_OPERATORS: ClassVar[dict] = {
            **Snowflake.Parser.COLUMN_OPERATORS,
            TokenType.COLON: lambda self, this, path: self._json_column_op(this, path),
        }

        TIMESTAMPS: ClassVar[dict] = Snowflake.Parser.TIMESTAMPS.copy() - {TokenType.TIME}

        RANGE_PARSERS: ClassVar[dict] = {
            **Snowflake.Parser.RANGE_PARSERS,
        }

        ALTER_PARSERS: ClassVar[dict] = {**Snowflake.Parser.ALTER_PARSERS}

        def _parse_list_agg(self) -> exp.GroupConcat:
            if self._match(TokenType.DISTINCT):
                args: list[exp.Expression | None] = [
                    self.expression(exp.Distinct, expressions=[self._parse_conjunction()])
                ]
                if self._match(TokenType.COMMA):
                    args.extend(self._parse_csv(self._parse_conjunction))
            else:
                args = self._parse_csv(self._parse_conjunction)

            return self.expression(exp.GroupConcat, this=args[0], separator=seq_get(args, 1))

        def _parse_types(
            self, *, check_func: bool = False, schema: bool = False, allow_identifiers: bool = True
        ) -> t.Optional[exp.Expression]:  # noqa: UP007
            this = super()._parse_types(check_func=check_func, schema=schema, allow_identifiers=allow_identifiers)
            # https://docs.snowflake.com/en/sql-reference/data-types-numeric Numeric datatype alias
            if (
                isinstance(this, exp.DataType)
                and this.is_type("numeric", "decimal", "number", "integer", "int", "smallint", "bigint")
                and not this.expressions
            ):
                return exp.DataType.build("DECIMAL(38,0)")
            return this

        def _parse_parameter(self) -> local_expression.Parameter:
            wrapped = self._match(TokenType.L_BRACE)
            this = self._parse_var() or self._parse_identifier() or self._parse_primary()
            self._match(TokenType.R_BRACE)
            suffix = ""
            if not self._match(TokenType.SPACE) or self._match(TokenType.DOT):
                suffix = self._parse_var() or self._parse_identifier() or self._parse_primary()

            return self.expression(local_expression.Parameter, this=this, wrapped=wrapped, suffix=suffix)

        def _get_table_alias(self):
            """
            :returns the `table alias` by looping through all the tokens until it finds the `From` token.
            Example:
            * SELECT .... FROM persons p => returns `p`
            * SELECT
                 ....
              FROM
                 dwh.vw_replacement_customer  d  => returns `d`
            """
            # Create a deep copy of Parser to avoid any modifications to original one
            self_copy = copy.deepcopy(self)
            found_from = True if self_copy._match(TokenType.FROM, advance=False) else False
            run = True
            indx = 0
            table_alias = None

            try:
                # iterate through all the tokens if tokens are available
                while self_copy._index < len(self_copy._tokens) and (run or found_from):
                    self_copy._advance()
                    indx += 1
                    found_from = True if self_copy._match(TokenType.FROM, advance=False) else False
                    # stop iterating when all the tokens are parsed
                    if indx == len(self_copy._tokens):
                        run = False
                    # get the table alias when FROM token is found
                    if found_from:
                        """
                        advance to two tokens after FROM, if there are available.
                        Handles (SELECT .... FROM persons p => returns `p`)
                        """
                        if self_copy._index + 2 < len(self_copy._tokens):
                            self_copy._advance(2)
                            table_alias = self_copy._curr.text
                            """
                          * if the table is of format :: `<DB>.<TABLE>` <TABLE_ALIAS>, advance to two more tokens
                            - Handles (`SELECT .... FROM dwh.vw_replacement_customer  d`  => returns d)
                          * if the table is of format :: `<DB>.<SCHEMA>.<TABLE>` <TABLE_ALIAS>, advance to 2 more tokens
                            - Handles (`SELECT .... FROM prod.public.tax_transact tt`  => returns tt)
                            """
                            while table_alias == "." and self_copy._index + 2 < len(self_copy._tokens):
                                self_copy._advance(2)
                                table_alias = self_copy._curr.text
                            break
            except Exception:
                logger.exception("Error while getting table alias.")

            return table_alias

        def _json_column_op(self, this, path):
            """
            Get the `table alias` using _get_table_alias() and it is used to check whether
            to remove `.value` from `<COL>.value`. We need to remove `.value` only if it refers
            to `Lateral View` alias.
            :return: the expression based on the alias.
            """
            table_alias = self._get_table_alias()
            is_name_value = this.name.upper() == "VALUE"
            is_path_value = path.alias_or_name.upper() == "VALUE"

            if isinstance(this, exp.Column) and this.table:
                col_table_alias = this.table.upper()
            elif isinstance(this, local_expression.Bracket) and isinstance(this.this, exp.Column) and this.this.table:
                col_table_alias = this.this.table.upper()
            else:
                col_table_alias = this.name.upper()

            is_table_alias = col_table_alias == table_alias.upper() if table_alias else False

            if not isinstance(this, exp.Bracket) and is_name_value:
                # If the column is referring to `lateral alias`, remove `.value` reference (not needed in Databricks)
                if table_alias and this.table != table_alias:
                    return self.expression(local_expression.Bracket, this=this.table, expressions=[path])
                """
                    if it is referring to `table_alias`, we need to keep `.value`. See below example:
                    - SELECT f.first, p.c.value.first, p.value FROM persons_struct AS p
                        LATERAL VIEW EXPLODE($p.$c.contact) AS f
                """
                return self.expression(local_expression.Bracket, this=this, expressions=[path])
            elif isinstance(this, local_expression.Bracket) and (is_name_value or is_table_alias):
                return self.expression(local_expression.Bracket, this=this, expressions=[path])
            elif isinstance(path, exp.Literal) and (path or is_path_value):
                return self.expression(local_expression.Bracket, this=this, expressions=[path])
            else:
                return self.expression(exp.Bracket, this=this, expressions=[path])



import logging
import os
from pathlib import Path

from databricks.labs.remorph.config import MorphConfig
from databricks.labs.remorph.helpers.execution_time import timeit
from databricks.labs.remorph.helpers.file_utils import (
    dir_walk,
    is_sql_file,
    make_dir,
    remove_bom,
)
from databricks.labs.remorph.helpers.morph_status import MorphStatus, ValidationError
from databricks.labs.remorph.helpers.validate import Validate
from databricks.labs.remorph.snow import dialect_utils
from databricks.labs.remorph.snow.sql_transpiler import SQLTranspiler

logger = logging.getLogger(__name__)


def process_file(config: MorphConfig, input_file: str | Path, output_file: str | Path):
    source = config.source
    skip_validation = config.skip_validation

    parse_error_list = []
    validate_error_list = []
    no_of_sqls = 0

    input_file = Path(input_file)
    output_file = Path(output_file)

    with input_file.open() as f:
        sql = remove_bom(f.read())

    lca_error = dialect_utils.check_for_unsupported_lca(source, sql, str(input_file))
    if lca_error:
        validate_error_list.append(lca_error)
    transpiler = SQLTranspiler(source, sql, str(input_file), parse_error_list)
    transpiled_sql = transpiler.transpile()

    with output_file.open("w") as w:
        for output in transpiled_sql:
            if output:
                no_of_sqls = no_of_sqls + 1
                if skip_validation:
                    w.write(output)
                    w.write("\n;\n")
                else:
                    validate = Validate(config.sdk_config)
                    output_string, exception = validate.validate_format_result(config, output)
                    w.write(output_string)
                    if exception is not None:
                        validate_error_list.append(ValidationError(str(input_file), exception))
            else:
                warning_message = (
                    f"Skipped a query from file {input_file!s}. "
                    f"Check for unsupported operations related to STREAM, TASK, SESSION etc."
                )
                logger.warning(warning_message)

    return no_of_sqls, parse_error_list, validate_error_list


def process_directory(config: MorphConfig, root: str | Path, base_root: str, files: list[str]):
    output_folder = config.output_folder
    parse_error_list = []
    validate_error_list = []
    counter = 0

    root = Path(root)

    for file in files:
        logger.debug(f"Processing file :{file}")
        if is_sql_file(file):
            if output_folder in (None, "None"):
                output_folder_base = root / "transpiled"
            else:
                output_folder_base = f'{output_folder.rstrip("/")}/{base_root}'

            output_file_name = Path(output_folder_base) / Path(file).name
            make_dir(output_folder_base)

            no_of_sqls, parse_error, validation_error = process_file(config, file, output_file_name)
            counter = counter + no_of_sqls
            parse_error_list.extend(parse_error)
            validate_error_list.extend(validation_error)
        else:
            # Only SQL files are processed with extension .sql or .ddl
            pass

    return counter, parse_error_list, validate_error_list


def process_recursive_dirs(config: MorphConfig):
    input_sql = Path(config.input_sql)
    parse_error_list = []
    validate_error_list = []

    file_list = []
    counter = 0
    for root, _, files in dir_walk(input_sql):
        base_root = str(root).replace(str(input_sql), "")
        folder = str(input_sql.resolve().joinpath(base_root))
        msg = f"Processing for sqls under this folder: {folder}"
        logger.info(msg)
        file_list.extend(files)
        no_of_sqls, parse_error, validation_error = process_directory(config, root, base_root, files)
        counter = counter + no_of_sqls
        parse_error_list.extend(parse_error)
        validate_error_list.extend(validation_error)

    error_log = parse_error_list + validate_error_list

    return MorphStatus(file_list, counter, len(parse_error_list), len(validate_error_list), error_log)


@timeit
def morph(config: MorphConfig):
    """
    Transpiles the SQL queries from one dialect to another.

    :param config: The configuration for the morph operation.
    """
    input_sql = Path(config.input_sql)
    skip_validation = config.skip_validation
    status = []
    result = MorphStatus([], 0, 0, 0, [])

    if input_sql.is_file():
        if is_sql_file(input_sql):
            msg = f"Processing for sqls under this file: {input_sql}"
            logger.info(msg)
            if config.output_folder in (None, "None"):
                output_folder = input_sql.parent / "transpiled"
            else:
                output_folder = Path(config.output_folder.rstrip("/"))

            make_dir(output_folder)
            output_file = output_folder / input_sql.name
            no_of_sqls, parse_error, validation_error = process_file(config, input_sql, output_file)
            error_log = parse_error + validation_error
            result = MorphStatus([str(input_sql)], no_of_sqls, len(parse_error), len(validation_error), error_log)
        else:
            msg = f"{input_sql} is not a SQL file."
            logger.warning(msg)
    elif input_sql.is_dir():
        result = process_recursive_dirs(config)
    else:
        msg = f"{input_sql} does not exist."
        logger.error(msg)
        raise FileNotFoundError(msg)

    parse_error_count = result.parse_error_count
    validate_error_count = result.validate_error_count

    error_list_count = parse_error_count + validate_error_count

    if not skip_validation:
        logger.info(f"No of Sql Failed while Validating: {validate_error_count}")

    if error_list_count > 0:
        error_log_file = Path.cwd() / f"err_{os.getpid()}.lst"
        with error_log_file.open("a") as e:
            e.writelines(f"{err}\n" for err in result.error_log_list)
    else:
        error_log_file = "None"

    status.append(
        {
            "total_files_processed": len(result.file_list),
            "total_queries_processed": result.no_of_queries,
            "no_of_sql_failed_while_parsing": parse_error_count,
            "no_of_sql_failed_while_validating": validate_error_count,
            "error_log_file": str(error_log_file),
        }
    )
    return status



import logging

from sqlglot import ErrorLevel, exp, parse
from sqlglot.expressions import Expression, Select

from databricks.labs.remorph.helpers.morph_status import ValidationError
from databricks.labs.remorph.snow.snowflake import Snow

logger = logging.getLogger(__name__)


def find_windows_in_select(select: Select) -> list[exp.Window]:
    window_expressions = []
    for expr in select.expressions:
        window_expr = expr.find(exp.Window)
        if window_expr:
            window_expressions.append(window_expr)
    return window_expressions


def check_for_unsupported_lca(
    dialect: str,
    sql: str,
    filename: str,
) -> ValidationError | None:
    """
    Check for presence of unsupported lateral column aliases in window expressions and where clauses
    :return: An error if found
    """
    if dialect.upper() == "SNOWFLAKE":
        dialect = Snow

    try:
        parsed_expr = parse(sql, read=dialect, error_level=ErrorLevel.RAISE)
    except Exception as e:
        logger.warning(f"Error while preprocessing {filename}: {e}")
        return None

    aliases_in_where = set()
    aliases_in_window = set()

    for expr in parsed_expr:
        if expr:
            for select in expr.find_all(exp.Select, bfs=False):
                alias_names = {alias.output_name for alias in select.expressions if isinstance(alias, exp.Alias)}
                aliases_in_where.update(_find_invalid_lca_in_where(select, alias_names))
                aliases_in_window.update(_find_invalid_lca_in_window(select, alias_names))

    if not (aliases_in_where or aliases_in_window):
        return None

    err_messages = [f"Unsupported operation found in file {filename}. Needs manual review of transpiled query."]
    if aliases_in_where:
        err_messages.append(f"Lateral column aliases `{', '.join(aliases_in_where)}` found in where clause.")

    if aliases_in_window:
        err_messages.append(f"Lateral column aliases `{', '.join(aliases_in_window)}` found in window expressions.")

    return ValidationError(filename, " ".join(err_messages))


def _find_invalid_lca_in_where(
    select_expr: Select,
    aliases: set[str],
) -> set[str]:
    aliases_in_where = set()
    where_ast: Expression = select_expr.args.get("where")
    if where_ast:
        for column in where_ast.find_all(exp.Column):
            if column.name in aliases:
                aliases_in_where.add(column.name)

    return aliases_in_where


def _find_invalid_lca_in_window(
    select_expr: Select,
    aliases: set[str],
) -> set[str]:
    aliases_in_window = set()
    windows = find_windows_in_select(select_expr)
    for window in windows:
        for column in window.find_all(exp.Column):
            if column.name in aliases:
                aliases_in_window.add(column.name)

    return aliases_in_window



from dataclasses import dataclass
from typing import ClassVar

from sqlglot import expressions as exp
from sqlglot.expressions import AggFunc, Condition, Expression, Func


class NthValue(AggFunc):
    arg_types: ClassVar[dict] = {"this": True, "offset": False}


class Parameter(Expression):
    arg_types: ClassVar[dict] = {"this": True, "wrapped": False, "suffix": False}


class Collate(Func):
    arg_types: ClassVar[dict] = {"this": True, "expressions": True}


class Bracket(Condition):
    arg_types: ClassVar[dict] = {"this": True, "expressions": True}


class Split(Func):
    """
    Redefined Split(sqlglot/expression) class with expression: False to handle default delimiter
    Please refer the test case `test_strtok_to_array` -> `select STRTOK_TO_ARRAY('my text is divided')`
    """

    arg_types: ClassVar[dict] = {"this": True, "expression": False, "limit": False}


class MakeDate(Func):
    arg_types: ClassVar[dict] = {"this": True, "expression": False, "zone": False}


class ConvertTimeZone(Func):
    arg_types: ClassVar[dict] = {"srcTZ": True, "tgtTZ": True, "this": False}


class TryToDate(Func):
    arg_types: ClassVar[dict] = {"this": True, "format": False}


class SplitPart(Func):
    arg_types: ClassVar[dict] = {"this": True, "expression": False, "partNum": False}


class StrTok(Func):
    arg_types: ClassVar[dict] = {"this": True, "expression": False, "partNum": False}


class TryToNumber(Func):
    arg_types: ClassVar[dict] = {"this": True, "expression": True, "precision": False, "scale": False}
    _sql_names: ClassVar[dict] = ["TRY_TO_DECIMAL", "TRY_TO_NUMBER", "TRY_TO_NUMERIC"]


class DateFormat(Func):
    arg_types: ClassVar[dict] = {"this": True, "expression": False}


class IsInteger(Func):
    pass


class JsonExtractPathText(Func):
    arg_types: ClassVar[dict] = {"this": True, "path_name": True}


class BitOr(AggFunc):
    pass


class ArrayConstructCompact(Func):
    arg_types: ClassVar[dict] = {"expressions": False}
    is_var_len_args = True


class ArrayIntersection(Func):
    arg_types: ClassVar[dict] = {"this": True, "expression": True}


class ArraySlice(Func):
    arg_types: ClassVar[dict] = {"this": True, "from": True, "to": True}


class ObjectKeys(Func):
    arg_types: ClassVar[dict] = {"this": True}


class ToBoolean(Func):
    arg_types: ClassVar[dict] = {"this": True, "raise_error": False}


class ToDouble(Func):
    pass


class ToObject(Func):
    pass


class ToNumber(Func):
    arg_types: ClassVar[dict] = {"this": True, "expression": False, "precision": False, "scale": False}
    _sql_names: ClassVar[list] = ["TO_DECIMAL", "TO_NUMBER", "TO_NUMERIC"]


class TimestampFromParts(Func):
    arg_types: ClassVar[dict] = {
        "this": True,
        "expression": True,
        "day": True,
        "hour": True,
        "min": True,
        "sec": True,
        "nanosec": False,
        "Zone": False,
    }


class ToVariant(Func):
    pass


class UUID(Func):
    arg_types: ClassVar[dict] = {"this": False, "name": False}


class DateTrunc(Func):
    arg_types: ClassVar[dict] = {"unit": False, "this": True, "zone": False}


@dataclass
class WithinGroupParams:
    agg_col: exp.Column
    order_col: exp.Column
    is_order_asc: bool



import json
import os

from databricks.labs.blueprint.cli import App
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.sdk import WorkspaceClient

from databricks.labs.remorph.config import MorphConfig
from databricks.labs.remorph.reconcile.execute import recon
from databricks.labs.remorph.transpiler.execute import morph

remorph = App(__file__)
logger = get_logger(__file__)


def raise_validation_exception(msg: str) -> Exception:
    raise Exception(msg)


@remorph.command
def transpile(
    w: WorkspaceClient,
    source: str,
    input_sql: str,
    output_folder: str,
    skip_validation: str,
    catalog_name: str,
    schema_name: str,
):
    """transpiles source dialect to databricks dialect"""
    logger.info(f"user: {w.current_user.me()}")

    if source.lower() not in ("snowflake", "tsql"):
        raise_validation_exception(
            f"Error: Invalid value for '--source': '{source}' is not one of 'snowflake', 'tsql'. "
        )
    if not os.path.exists(input_sql) or input_sql in (None, ""):
        raise_validation_exception(f"Error: Invalid value for '--input_sql': Path '{input_sql}' does not exist.")
    if output_folder == "":
        output_folder = None
    if skip_validation.lower() not in ("true", "false"):
        raise_validation_exception(
            f"Error: Invalid value for '--skip_validation': '{skip_validation}' is not one of 'true', 'false'. "
        )

    config = MorphConfig(
        source=source.lower(),
        input_sql=input_sql,
        output_folder=output_folder,
        skip_validation=skip_validation.lower() == "true",  # convert to bool
        catalog_name=catalog_name,
        schema_name=schema_name,
        sdk_config=w.config,
    )

    status = morph(config)

    print(json.dumps(status))


@remorph.command
def reconcile(w: WorkspaceClient, recon_conf: str, conn_profile: str, source: str, report: str):
    """reconciles source to databricks datasets"""
    logger.info(f"user: {w.current_user.me()}")
    if not os.path.exists(recon_conf) or recon_conf in (None, ""):
        raise_validation_exception(f"Error: Invalid value for '--recon_conf': Path '{recon_conf}' does not exist.")
    if not os.path.exists(conn_profile) or conn_profile in (None, ""):
        raise_validation_exception(f"Error: Invalid value for '--conn_profile': Path '{conn_profile}' does not exist.")
    if source.lower() not in "snowflake":
        raise_validation_exception(f"Error: Invalid value for '--source': '{source}' is not one of 'snowflake'. ")
    if report.lower() not in ("data", "schema", "all"):
        raise_validation_exception(
            f"Error: Invalid value for '--report': '{report}' is not one of 'data', 'schema', 'all' "
        )

    recon(recon_conf, conn_profile, source, report)


if __name__ == "__main__":
    remorph()



from dataclasses import dataclass


@dataclass
class ParseError:
    file_name: str
    exception: str


@dataclass
class ValidationError:
    file_name: str
    exception: str


@dataclass
class MorphStatus:
    file_list: list[str]
    no_of_queries: int
    parse_error_count: int
    validate_error_count: int
    error_log_list: list[ParseError | ValidationError] | None



import logging
import re
from typing import ClassVar

from sqlglot import expressions as exp
from sqlglot.dialects import hive
from sqlglot.dialects.databricks import Databricks
from sqlglot.dialects.dialect import rename_func
from sqlglot.errors import ParseError, UnsupportedError
from sqlglot.helper import apply_index_offset, csv

from databricks.labs.remorph.snow import local_expression

logger = logging.getLogger(__name__)

VALID_DATABRICKS_TYPES = {
    "BIGINT",
    "BINARY",
    "BOOLEAN",
    "DATE",
    "DECIMAL",
    "DOUBLE",
    "FLOAT",
    "INT",
    "INTERVAL",
    "VOID",
    "SMALLINT",
    "STRING",
    "TIMESTAMP",
    "TINYINT",
    "ARRAY",
    "MAP",
    "STRUCT",
}


def timestamptrunc_sql(self, expression: exp.TimestampTrunc) -> str:
    return self.func("DATE_TRUNC", exp.Literal.string(expression.text("unit").upper()), self.sql(expression.this))


def _parm_sfx(self, expression: local_expression.Parameter) -> str:
    this = self.sql(expression, "this")
    this = f"{{{this}}}" if expression.args.get("wrapped") else f"{this}"
    suffix = self.sql(expression, "suffix")
    PARAMETER_TOKEN = "$"  # noqa: N806
    return f"{PARAMETER_TOKEN}{this}{suffix}"


def _lateral_bracket_sql(self, expression: local_expression.Bracket) -> str:
    """Overwrites `sqlglot/generator.py` `bracket_sql()` function
    to convert <TABLE_ALIAS>`[COL_NAME]` to <TABLE_ALIAS>`.COL_NAME`.
    Example: c[val] ==> c.val
    """
    expressions = apply_index_offset(expression.this, expression.expressions, self.dialect.INDEX_OFFSET)
    expressions = [self.sql(e.alias_or_name.strip("'")) for e in expressions]
    # If expression contains space in between encode it in backticks(``):
    # e.g. ref."ID Number" -> ref.`ID Number`.
    expressions_sql = ", ".join(f"`{e}`" if " " in e else e for e in expressions)
    return f"{self.sql(expression, 'this')}.{expressions_sql}"


def _format_create_sql(self, expression: exp.Create) -> str:
    expression = expression.copy()

    # Remove modifiers in order to simplify the schema.  For example, this removes things like "IF NOT EXISTS"
    # from "CREATE TABLE foo IF NOT EXISTS".
    args_to_delete = ["temporary", "transient", "external", "replace", "exists", "unique", "materialized", "properties"]
    for arg_to_delete in args_to_delete:
        if expression.args.get(arg_to_delete):
            del expression.args[arg_to_delete]

    return hive._create_sql(self, expression)


def _curr_time():
    return "date_format(current_timestamp(), 'HH:mm:ss')"


def _lateral_view(self, expression: exp.Lateral) -> str:
    str_lateral_view = "LATERAL VIEW"
    str_outer = "OUTER"
    str_explode = "EXPLODE("
    str_pfx = f"{str_lateral_view} {str_explode}"
    str_alias = ")"

    for expr, _, _ in expression.walk(bfs=True, prune=lambda *_: False):
        match expr:
            case exp.Explode():
                if expr.key.upper() != "EXPLODE":
                    continue
                for node, _, _ in expr.walk(bfs=True, prune=lambda *_: False):
                    if not isinstance(node, exp.Kwarg):
                        continue
                    if not isinstance(node.this, exp.Var):
                        continue

                    node_name = str(node.this).upper()
                    match node_name:
                        case "INPUT":
                            # Added if block to handle Dynamic variables `${}`
                            node_expr = f"{node.expression}".replace("@", "$")
                            if "PARSE_JSON" in node_expr:
                                node_expr = node_expr.replace("PARSE_JSON", "FROM_JSON")
                                msg = (
                                    f"***Warning***: you need to explicitly specify "
                                    f"`SCHEMA` for column(s) in `{node_expr}`"
                                )
                                logger.warning(msg)
                            str_pfx = str_pfx + node_expr
                        case "PATH":
                            str_pfx = str_pfx + f".{node.expression}".replace("'", "").replace('"', "`")
                        case "OUTER":
                            str_pfx = str_pfx.replace(str_lateral_view, f"{str_lateral_view} {str_outer}")
                            # [TODO]: Implement for options: RECURSIVE and MODE
            case exp.TableAlias():
                str_alias = str_alias + f" AS {expr.name}"

    return self.sql(str_pfx + str_alias)


# [TODO] Add more datatype coverage https://docs.databricks.com/sql/language-manual/sql-ref-datatypes.html
def _datatype_map(self, expression) -> str:
    if expression.this in [exp.DataType.Type.VARCHAR, exp.DataType.Type.NVARCHAR, exp.DataType.Type.CHAR]:
        return "STRING"
    if expression.this in [exp.DataType.Type.TIMESTAMP, exp.DataType.Type.TIMESTAMPLTZ]:
        return "TIMESTAMP"
    return self.datatype_sql(expression)


def try_to_date(self, expression: local_expression.TryToDate):
    func = "TRY_TO_TIMESTAMP"
    time_format = self.sql(expression, "format")
    if not time_format:
        time_format = hive.Hive.DATE_FORMAT

    ts_result = self.func(func, expression.this, time_format)
    return exp.Date(this=ts_result)


def try_to_number(self, expression: local_expression.TryToNumber):
    func = "TRY_TO_NUMBER"
    precision = self.sql(expression, "precision")
    scale = self.sql(expression, "scale")

    if not precision:
        precision = 38

    if not scale:
        scale = 0

    func_expr = self.func(func, expression.this)
    if expression.expression:
        func_expr = self.func(func, expression.this, expression.expression)

    return f"CAST({func_expr} AS DECIMAL({precision}, {scale}))"


def _to_boolean(self: Databricks.Generator, expression: local_expression.ToBoolean) -> str:
    this = self.sql(expression, "this")
    logger.debug(f"Converting {this} to Boolean")
    raise_error = self.sql(expression, "raise_error")
    raise_error_str = "RAISE_ERROR('Invalid parameter type for TO_BOOLEAN')" if bool(int(raise_error)) else "NULL"
    transformed = f"""
    CASE
       WHEN {this} IS NULL THEN NULL
       WHEN TYPEOF({this}) = 'boolean' THEN BOOLEAN({this})
       WHEN TYPEOF({this}) = 'string' THEN
           CASE
               WHEN LOWER({this}) IN ('true', 't', 'yes', 'y', 'on', '1') THEN TRUE
               WHEN LOWER({this}) IN ('false', 'f', 'no', 'n', 'off', '0') THEN FALSE
               ELSE RAISE_ERROR('Boolean value of x is not recognized by TO_BOOLEAN')
               END
       WHEN ISNOTNULL(TRY_CAST({this} AS DOUBLE)) THEN
           CASE
               WHEN ISNAN(CAST({this} AS DOUBLE)) OR CAST({this} AS DOUBLE) = DOUBLE('infinity') THEN
                    RAISE_ERROR('Invalid parameter type for TO_BOOLEAN')
               ELSE CAST({this} AS DOUBLE) != 0.0
               END
       ELSE {raise_error_str}
       END
    """
    return transformed


def _is_integer(self: Databricks.Generator, expression: local_expression.IsInteger) -> str:
    this = self.sql(expression, "this")
    transformed = f"""
    CASE
       WHEN {this} IS NULL THEN NULL
       WHEN {this} RLIKE '^-?[0-9]+$' AND TRY_CAST({this} AS INT) IS NOT NULL THEN TRUE
       ELSE FALSE
       END
    """
    return transformed


def _parse_json_extract_path_text(self: Databricks.Generator, expression: local_expression.JsonExtractPathText) -> str:
    this = self.sql(expression, "this")
    path_name = expression.args["path_name"]
    if path_name.is_string:
        path = f"{self.dialect.QUOTE_START}$.{expression.text('path_name')}{self.dialect.QUOTE_END}"
    else:
        path = f"CONCAT('$.', {self.sql(expression, 'path_name')})"
    return f"GET_JSON_OBJECT({this}, {path})"


def _array_construct_compact(self: Databricks.Generator, expression: local_expression.ArrayConstructCompact) -> str:
    exclude = "ARRAY(NULL)"
    array_expr = f"ARRAY({self.expressions(expression, flat=True)})"
    return f"ARRAY_EXCEPT({array_expr}, {exclude})"


def _array_slice(self: Databricks.Generator, expression: local_expression.ArraySlice) -> str:
    from_expr = self.sql(expression, "from")
    # In Databricks: array indices start at 1 in function `slice(array, start, length)`
    from_expr = 1 if from_expr == "0" else from_expr

    to_expr = self.sql(expression, "to")
    # Convert string expression to number and check if it is negative number
    if int(to_expr) < 0:
        err_message = "In Databricks: function `slice` length must be greater than or equal to 0"
        raise UnsupportedError(err_message)

    func = "SLICE"
    func_expr = self.func(func, expression.this, exp.Literal.number(from_expr), expression.args["to"])
    return func_expr


def _parse_json(self, expr: exp.ParseJSON):
    """
    Converts `PARSE_JSON` function to `FROM_JSON` function.
    Schema is a mandatory argument for Databricks `FROM_JSON` function
    [FROM_JSON](https://docs.databricks.com/en/sql/language-manual/functions/from_json.html)
    Need to explicitly specify the Schema {<COL_NAME>_SCHEMA} in the current execution environment
    """
    expr_this = self.sql(expr, "this")
    column = expr_this.replace("'", "").upper()
    conv_expr = self.func("FROM_JSON", expr_this, f"{{{column}_SCHEMA}}")
    warning_msg = (
        f"***Warning***: you need to explicitly specify `SCHEMA` for `{column}` column in expression: `{conv_expr}`"
    )
    logger.warning(warning_msg)
    return conv_expr


def _to_number(self, expression: local_expression.TryToNumber):
    func = "TO_NUMBER"
    precision = self.sql(expression, "precision")
    scale = self.sql(expression, "scale")

    if not precision:
        precision = 38

    if not scale:
        scale = 0

    func_expr = self.func(func, expression.this)
    if expression.expression:
        func_expr = self.func(func, expression.this, expression.expression)
    else:
        exception_msg = f"""Error Parsing expression {expression}:
                         * `format`: is required in Databricks [mandatory]
                         * `precision` and `scale`: are considered as (38, 0) if not specified.
                      """
        raise UnsupportedError(exception_msg)

    return f"CAST({func_expr} AS DECIMAL({precision}, {scale}))"


def _uuid(self: Databricks.Generator, expression: local_expression.UUID) -> str:
    namespace = self.sql(expression, "this")
    name = self.sql(expression, "name")

    if namespace and name:
        logger.warning("UUID version 5 is not supported currently. Needs manual intervention.")
        return f"UUID({namespace}, {name})"
    else:
        return "UUID()"


def _parse_date_trunc(self: Databricks.Generator, expression: local_expression.DateTrunc) -> str:
    if not expression.args.get("unit"):
        error_message = f"Required keyword: 'unit' missing for {exp.DateTrunc}"
        raise UnsupportedError(error_message)
    return self.func("TRUNC", expression.this, expression.args.get("unit"))


def _get_within_group_params(
    expr: exp.ArrayAgg | exp.GroupConcat,
    within_group: exp.WithinGroup,
) -> local_expression.WithinGroupParams:
    has_distinct = isinstance(expr.this, exp.Distinct)
    agg_col = expr.this.expressions[0] if has_distinct else expr.this
    order_expr = within_group.expression
    order_col = order_expr.expressions[0].this
    desc = order_expr.expressions[0].args.get("desc")
    is_order_asc = not desc or exp.false() == desc
    # In Snow, if both DISTINCT and WITHIN GROUP are specified, both must refer to the same column.
    # Ref: https://docs.snowflake.com/en/sql-reference/functions/array_agg#usage-notes
    # TODO: Check the same restriction applies for other source dialects to be added in the future
    if has_distinct and agg_col != order_col:
        raise ParseError("If both DISTINCT and WITHIN GROUP are specified, both must refer to the same column.")
    return local_expression.WithinGroupParams(
        agg_col=agg_col,
        order_col=order_col,
        is_order_asc=is_order_asc,
    )


def _create_named_struct_for_cmp(agg_col, order_col) -> exp.Expression:
    named_struct_func = exp.Anonymous(
        this="named_struct",
        expressions=[
            exp.Literal(this="value", is_string=True),
            agg_col,
            exp.Literal(this="sort_by", is_string=True),
            order_col,
        ],
    )
    return named_struct_func


class Databricks(Databricks):
    # Instantiate Databricks Dialect
    databricks = Databricks()

    class Generator(databricks.Generator):
        COLLATE_IS_FUNC = True
        # [TODO]: Variant needs to be transformed better, for now parsing to string was deemed as the choice.
        TYPE_MAPPING: ClassVar[dict] = {
            **Databricks.Generator.TYPE_MAPPING,
            exp.DataType.Type.TINYINT: "TINYINT",
            exp.DataType.Type.SMALLINT: "SMALLINT",
            exp.DataType.Type.INT: "INT",
            exp.DataType.Type.BIGINT: "BIGINT",
            exp.DataType.Type.DATETIME: "TIMESTAMP",
            exp.DataType.Type.VARCHAR: "STRING",
            exp.DataType.Type.VARIANT: "STRING",
            exp.DataType.Type.FLOAT: "DOUBLE",
        }

        TRANSFORMS: ClassVar[dict] = {
            **Databricks.Generator.TRANSFORMS,
            exp.Create: _format_create_sql,
            exp.DataType: _datatype_map,
            exp.CurrentTime: _curr_time(),
            exp.Lateral: _lateral_view,
            exp.FromBase64: rename_func("UNBASE64"),
            local_expression.Parameter: _parm_sfx,
            local_expression.ToBoolean: _to_boolean,
            local_expression.Bracket: _lateral_bracket_sql,
            local_expression.MakeDate: rename_func("MAKE_DATE"),
            local_expression.TryToDate: try_to_date,
            local_expression.TryToNumber: try_to_number,
            local_expression.IsInteger: _is_integer,
            local_expression.JsonExtractPathText: _parse_json_extract_path_text,
            local_expression.BitOr: rename_func("BIT_OR"),
            local_expression.ArrayConstructCompact: _array_construct_compact,
            local_expression.ArrayIntersection: rename_func("ARRAY_INTERSECT"),
            local_expression.ArraySlice: _array_slice,
            local_expression.ObjectKeys: rename_func("JSON_OBJECT_KEYS"),
            exp.ParseJSON: _parse_json,
            local_expression.TimestampFromParts: rename_func("MAKE_TIMESTAMP"),
            local_expression.ToDouble: rename_func("DOUBLE"),
            exp.Rand: rename_func("RANDOM"),
            local_expression.ToVariant: rename_func("TO_JSON"),
            local_expression.ToObject: rename_func("TO_JSON"),
            exp.ToBase64: rename_func("BASE64"),
            local_expression.ToNumber: _to_number,
            local_expression.UUID: _uuid,
            local_expression.DateTrunc: _parse_date_trunc,
            exp.ApproxQuantile: rename_func("APPROX_PERCENTILE"),
            exp.TimestampTrunc: timestamptrunc_sql,
        }

        def join_sql(self, expression: exp.Join) -> str:
            """Overwrites `join_sql()` in `sqlglot/generator.py`
            Added logic to handle Lateral View
            """
            op_list = [
                expression.method,
                "GLOBAL" if expression.args.get("global") else None,
                expression.side,
                expression.kind,
                expression.hint if self.JOIN_HINTS else None,
            ]

            op_sql = " ".join(op for op in op_list if op)
            on_sql = self.sql(expression, "on")
            using = expression.args.get("using")

            if not on_sql and using:
                on_sql = csv(*(self.sql(column) for column in using))

            this_sql = self.sql(expression, "this")

            if on_sql:
                on_sql = self.indent(on_sql, skip_first=True)
                space = self.seg(" " * self.pad) if self.pretty else " "
                if using:
                    on_sql = f"{space}USING ({on_sql})"
                else:
                    on_sql = f"{space}ON {on_sql}"
            # Added the below elif block to handle Lateral View clause
            elif not op_sql and isinstance(expression.this, exp.Lateral):
                return f"\n {this_sql}"
            elif not op_sql:
                return f", {this_sql}"

            op_sql = f"{op_sql} JOIN" if op_sql else "JOIN"
            return f"{self.seg(op_sql)} {this_sql}{on_sql}"

        def arrayagg_sql(self, expr: exp.ArrayAgg) -> str:
            sql = self.func("ARRAY_AGG", expr.this)
            within_group = expr.parent if isinstance(expr.parent, exp.WithinGroup) else None
            if not within_group:
                return sql

            wg_params = _get_within_group_params(expr, within_group)
            if wg_params.agg_col == wg_params.order_col:
                return f"SORT_ARRAY({sql}{'' if wg_params.is_order_asc else ', FALSE'})"

            named_struct_func = _create_named_struct_for_cmp(wg_params.agg_col, wg_params.order_col)
            array_sort = self.func(
                "ARRAY_SORT",
                self.func("ARRAY_AGG", named_struct_func),
                f"""(left, right) -> CASE
                        WHEN left.sort_by < right.sort_by THEN {'-1' if wg_params.is_order_asc else '1'}
                        WHEN left.sort_by > right.sort_by THEN {'1' if wg_params.is_order_asc else '-1'}
                        ELSE 0
                    END""",
            )
            return self.func("TRANSFORM", array_sort, "s -> s.value")

        def groupconcat_sql(self, expr: exp.GroupConcat) -> str:
            arr_agg = exp.ArrayAgg(this=expr.this)
            within_group = expr.parent.copy() if isinstance(expr.parent, exp.WithinGroup) else None
            if within_group:
                arr_agg.parent = within_group

            return self.func(
                "ARRAY_JOIN",
                arr_agg,
                expr.args.get("separator") or exp.Literal(this="", is_string=True),
            )

        def withingroup_sql(self, expression: exp.WithinGroup) -> str:
            agg_expr = expression.this
            if isinstance(agg_expr, (exp.ArrayAgg, exp.GroupConcat)):
                return self.sql(agg_expr)
            else:
                return super().withingroup_sql(expression)

        def split_sql(self, expression: local_expression.Split) -> str:
            """
            :param expression: local_expression.Split expression to be parsed
            :return: Converted expression (SPLIT) compatible with Databricks
            """
            delimiter = " "
            # To handle default delimiter
            if expression.expression:
                delimiter = expression.expression.name

            # Parsing logic to handle String and Table columns
            if expression.name and isinstance(expression.name, str):
                expr_name = f"'{expression.name}'"
            else:
                expr_name = expression.args["this"]
            return f"""SPLIT({expr_name},'[{delimiter}]')"""

        def delete_sql(self, expression: exp.Delete) -> str:
            this = self.sql(expression, "this")
            using = self.sql(expression, "using")
            where = self.sql(expression, "where")
            returning = self.sql(expression, "returning")
            limit = self.sql(expression, "limit")
            tables = self.expressions(expression, key="tables")
            tables = f" {tables}" if tables else ""

            if using:
                using = f" USING {using}" if using else ""
                where = where.replace("WHERE", "ON")
            else:
                this = f"FROM {this}" if this else ""

            if self.RETURNING_END:
                expression_sql = f" {this}{using}{where}{returning}{limit}"
            else:
                expression_sql = f"{returning}{this}{where}{limit}"

            if using:
                return self.prepend_ctes(expression, f"MERGE INTO {tables}{expression_sql} WHEN MATCHED THEN DELETE;")
            else:
                return self.prepend_ctes(expression, f"DELETE{tables}{expression_sql};")

        def converttimezone_sql(self, expression: local_expression.ConvertTimeZone):
            func = "CONVERT_TIMEZONE"
            expr = expression.args["tgtTZ"]
            if len(expression.args) == 3 and expression.args.get("this"):
                expr = expression.args["this"]

            expr = f"'{expr.name}'" if isinstance(expr, exp.Cast) else expr

            result = self.func(func, expression.args["srcTZ"], expr)
            if len(expression.args) == 3:
                result = self.func(func, expression.args["srcTZ"], expression.args["tgtTZ"], expr)

            return result

        def strtok_sql(self, expression: local_expression.StrTok) -> str:
            """
            :param expression: local_expression.StrTok expression to be parsed
            :return: Converted expression (SPLIT_PART) compatible with Databricks
            """
            # To handle default delimiter
            if expression.expression:
                delimiter = expression.expression.name
            else:
                delimiter = " "

            # Handle String and Table columns
            if expression.name and isinstance(expression.name, str):
                expr_name = f"'{expression.name}'"
            else:
                expr_name = expression.args["this"]

            # Handle Partition Number
            if len(expression.args) == 3 and expression.args.get("partNum"):
                part_num = expression.args["partNum"]
            else:
                part_num = 1

            return f"SPLIT_PART({expr_name}, '{delimiter}', {part_num})"

        def splitpart_sql(self, expression: local_expression.SplitPart) -> str:
            """
            :param expression: local_expression.SplitPart expression to be parsed
            :return: Converted expression (SPLIT_PART) compatible with Databricks
            """
            expr_name = self.sql(expression.this)
            delimiter = self.sql(expression.expression)
            part_num = self.sql(expression.args["partNum"])
            return f"SPLIT_PART({expr_name}, {delimiter}, {part_num})"

        def transaction_sql(self, expression: exp.Transaction) -> str:  # noqa: ARG002
            """
            Skip begin command
            :param expression:
            :return: Empty string for unsupported operation
            """
            return ""

        def rollback_sql(self, expression: exp.Rollback) -> str:  # noqa: ARG002
            """
            Skip rollback command
            :param expression:
            :return: Empty string for unsupported operation
            """
            return ""

        def commit_sql(self, expression: exp.Rollback) -> str:  # noqa: ARG002
            """
            Skip commit command
            :param expression:
            :return: Empty string for unsupported operation
            """
            return ""

        def command_sql(self, expression: exp.Command) -> str:
            """
            Skip any session, stream, task related commands
            :param expression:
            :return: Empty string for unsupported operations or objects
            """
            filtered_commands = [
                "CREATE",
                "ALTER",
                "DESCRIBE",
                "DROP",
                "SHOW",
                "EXECUTE",
            ]
            ignored_objects = [
                "STREAM",
                "TASK",
                "STREAMS",
                "TASKS",
                "SESSION",
            ]

            command = self.sql(expression, "this").upper()
            expr = expression.text("expression").strip()
            obj = re.split(r"\s+", expr, maxsplit=2)[0].upper() if expr else ""
            if command in filtered_commands and obj in ignored_objects:
                return ""
            return f"{command} {expr}"

        def currenttimestamp_sql(self, _: exp.CurrentTimestamp) -> str:
            return self.func("CURRENT_TIMESTAMP")



from sqlglot import transpile

from databricks.labs.remorph.helpers.morph_status import ParseError
from databricks.labs.remorph.snow.databricks import Databricks
from databricks.labs.remorph.snow.snowflake import Snow


class SQLTranspiler:
    def __init__(self, source: str, sql: str, file_nm: str, error_list: list[ParseError]):
        self.source = source
        self.sql = sql
        self.file_nm = file_nm
        self.error_list = error_list

    def transpile(self) -> str:
        if self.source.upper() == "SNOWFLAKE":
            dialect = Snow
        else:
            dialect = self.source.lower()

        try:
            transpiled_sql = transpile(self.sql, read=dialect, write=Databricks, pretty=True, error_level=None)
        except Exception as e:
            transpiled_sql = ""

            self.error_list.append(ParseError(self.file_nm, e))

        return transpiled_sql



import codecs
from pathlib import Path


# Optionally check to see if a string begins with a Byte Order Mark
# such a character will cause the transpiler to fail
def remove_bom(input_string: str) -> str:
    """
    Removes the Byte Order Mark (BOM) from the given string if it exists.
    :param input_string: String to remove BOM from
    :return: String without BOM
    """
    output_string = input_string

    # Check and remove UTF-16 (LE and BE) BOM
    if input_string.startswith(codecs.BOM_UTF16_BE.decode("utf-16-be")):
        output_string = input_string[len(codecs.BOM_UTF16_BE.decode("utf-16-be")) :]
    elif input_string.startswith(codecs.BOM_UTF16_LE.decode("utf-16-le")):
        output_string = input_string[len(codecs.BOM_UTF16_LE.decode("utf-16-le")) :]
    elif input_string.startswith(codecs.BOM_UTF16.decode("utf-16")):
        output_string = input_string[len(codecs.BOM_UTF16.decode("utf-16")) :]
    # Check and remove UTF-32 (LE and BE) BOM
    elif input_string.startswith(codecs.BOM_UTF32_BE.decode("utf-32-be")):
        output_string = input_string[len(codecs.BOM_UTF32_BE.decode("utf-32-be")) :]
    elif input_string.startswith(codecs.BOM_UTF32_LE.decode("utf-32-le")):
        output_string = input_string[len(codecs.BOM_UTF32_LE.decode("utf-32-le")) :]
    elif input_string.startswith(codecs.BOM_UTF32.decode("utf-32")):
        output_string = input_string[len(codecs.BOM_UTF32.decode("utf-32")) :]
    # Check and remove UTF-8 BOM
    elif input_string.startswith(codecs.BOM_UTF8.decode("utf-8")):
        output_string = input_string[len(codecs.BOM_UTF8.decode("utf-8")) :]

    return output_string


def is_sql_file(file: str | Path) -> bool:
    """
    Checks if the given file is a SQL file.

    :param file: The name of the file to check.
    :return: True if the file is a SQL file (i.e., its extension is either .sql or .ddl), False otherwise.
    """
    file_extension = Path(file).suffix
    return file_extension.lower() in [".sql", ".ddl"]


def make_dir(path: str | Path) -> None:
    """
    Creates a directory at the specified path if it does not already exist.

    :param path: The path where the directory should be created.
    """
    Path(path).mkdir(parents=True, exist_ok=True)


def dir_walk(root: Path):
    """
    Walks the directory tree rooted at the given path, yielding a tuple containing the root directory, a list of
    :param root: Path
    :return: tuple of  root, subdirectory , files
    """
    sub_dirs = [d for d in root.iterdir() if d.is_dir()]
    files = [f for f in root.iterdir() if f.is_file()]
    yield root, sub_dirs, files
    for s in sub_dirs:
        yield from dir_walk(s)



from dataclasses import dataclass

from databricks.labs.blueprint.entrypoint import get_logger
from databricks.sdk.core import Config

logger = get_logger(__file__)


@dataclass
class MorphConfig:
    source: str
    input_sql: str
    output_folder: str | None
    sdk_config: Config | None
    skip_validation: bool = False
    catalog_name: str = "transpiler_test"
    schema_name: str = "convertor_test"

    def __post_init__(self):
        # skip validation is false
        if not self.skip_validation:
            if self.sdk_config is None:
                logger.error("sdk config is required when skip_validation is false")
                raise ValueError("sdk config is required when skip_validation is false")
